import numpy as np
from scipy import interpolate
from drivecycle import utils
import inspect
from typing import List, Any, Optional
import numpy.typing as npt


def energy_model(traj: List[float],
                  elv: Optional[List[List[float]]] = None,
                  capacity: int = 300, #kWh
                  power_aux:float = 0.0, #auxiliary loads kWh
                  **kwargs: str) -> npt.NDArray[Any]:
    """Battery State of Charge (SoC) modelling.

    Args:
        traj (list): time, velocity, distance list
        elv (list): elevation along the route
        capacity (int): batter capacity (kWh)
        power_aux (float): auxiliary load on the vehicle (kWh)

    Returns:
        list: (n,5) numpy array of time, velocity, distance, cumulative power, SoC
    """

    data = np.c_[traj, np.zeros((len(traj), 2))]  # Power, SoC
    data[0,4]=1
    alpha = None

    if elv is not None:
        x = [i[0] for i in elv]
        y = [i[1] for i in elv]
        elv_f = interpolate.interp1d(x, y)

    for i, (prev, curr) in enumerate(utils.pairwise(traj)):

        t0, v0, d0 = prev
        t1, v1, d1 = curr

        accel = (v1 - v0) / (t1 - t0)

        if v1 == 0:
            data[i + 1, 3:5] = data[i, 3:5]
        else:
            if elv is not None:
                try:
                    alpha = (elv_f(d1) - elv_f(d0)) / (d1 - d0)
                except ValueError: # d1 or d0 is outside elv range, don't include alpha in calcluation
                    alpha = None
                
            # determine tractive force
            trac_force_kwargs = list(
                inspect.signature(tractive_force).parameters)
            force = tractive_force(
                v1,
                accel,
                alpha=alpha,
                **{
                    i: kwargs[i]  # type: ignore
                    for i in kwargs if i in trac_force_kwargs
                },
            )

            # power supplied to or from the battery
            batt_power_kwargs = list(
                inspect.signature(battery_power).parameters)
            power_batt = battery_power(
                v1, 
                force, 
                **{
                    i: kwargs[i]  # type: ignore
                    for i in kwargs if i in batt_power_kwargs
                },
            )

            # Include accessory/auxiliary power
            power = (power_batt/1000) + power_aux # Total power in kW

            Einst = power * ((t1-t0)/3600) # Instantaneous energy in kWh

            #print(f"Power: {power}, Einst: {Einst}, SOC: {Einst/capacity}")
            
            data[i + 1, 3] = data[i,3] + Einst
            data[i + 1, 4] = data[i,4] - (Einst/capacity)

    return data

def battery_model(traj: List[float],
                  elv: Optional[List[List[float]]] = None,
                  num_cells: int = 2000,
                  capacity: int = 75,
                  k: float = 1.0,
                  battery_type: str = "LI-ION",
                  power_aux: float = 0.0,
                  **kwargs: str) -> npt.NDArray[Any]:
    """Battery Depth of Discharge (DoD) modelling.

    Battery depth of discharge modelling using calculations for 
    current draw from the battery.

    Args:
        traj (list): time, velocity, distance list
        elv (list): elevation along the route
        num_cells (int): number of battery cells
        capacity (int): amp hour capacity of cell, 1C
        k (float): peuker coefficient
        battery_type (str): battery chemical type

    Returns:
        list: (n,6) numpy array of time, velocity, distance, power, current, DoD

    """

    data = np.c_[traj, np.zeros((len(traj), 3))]  # Power, Cr, DoD
    r_in = (0.022 / capacity) * num_cells  # Internal resistance
    peu_cap = (np.power((capacity / 1), k)) * 1
    alpha = None

    if elv is not None:
        x = [i[0] for i in elv]
        y = [i[1] for i in elv]
        elv_f = interpolate.interp1d(x, y)

    for i, (prev, curr) in enumerate(utils.pairwise(traj)):

        t0, v0, d0 = prev
        t1, v1, d1 = curr

        accel = (v1 - v0) / (t1 - t0)

        if v1 == 0:
            data[i + 1, 3:6] = data[i, 3:6]
        else:

            if elv is not None:
                alpha = (elv_f(d1) - elv_f(d0)) / (d1 - d0)

            trac_force_kwargs = list(
                inspect.signature(tractive_force).parameters)
            force = tractive_force(
                v1,
                accel,
                alpha=alpha,
                **{
                    i: kwargs[i]  # type: ignore
                    for i in kwargs if i in trac_force_kwargs
                },
            )

            batt_power_kwargs = list(
                inspect.signature(battery_power).parameters)
            power_batt = battery_power(
                v1,
                force,
                **{
                    i: kwargs[i]  # type: ignore
                    for i in kwargs if i in batt_power_kwargs
                })

            data[i + 1, 3] = power_batt+power_aux

            # determine open circuit voltage given battery type and number of cells
            voltage = open_circuit_voltage(data[i, 5],
                                           type=battery_type) * num_cells
                                           
            if power_batt > 0:
                a = np.power(voltage, 2) - (4 * r_in * power_batt)
                assert a > 0, "Cannot determine current, insufficient total voltage."

                # Current drawn from the battery
                batt_current = (voltage - np.sqrt(a)) / (2 * r_in)
                data[i + 1,
                     4] = data[i, 4] + ((batt_current*(t1-t0)) / 3600)
            elif power_batt == 0:
                batt_current = 0
            elif power_batt < 0:
                # Regenerative braking double the internal resistance.
                a = np.power(voltage, 2) + (4 * 1.0 * r_in * -power_batt)
                assert a > 0, "Cannot determine current, insufficient total voltage."

                # Current drawn from the battery
                batt_current = (-voltage + np.sqrt(a)) / (2 * 1.0 * r_in)
                data[i + 1, 4] = data[i, 4] - ((batt_current*(t1-t0)) / 3600)

            # assuming battery cells are in series therefor capacity is same as input
            data[i + 1, 5] = data[i + 1, 4] / peu_cap 

    return data  # type: ignore


def battery_power(
    v: float,
    force: float, # tractive force in N
    gear_ratio: float = 4.66,
    radius_wheel: float = 0.5, # wheel radius in m
    regen_ratio: float = 0.5, # proportion of braking forcing to use for regen
    trans_eff: float = 0.95,  # transmission efficiency
    motor_eff: float = 0.90, # motor efficiency
    pow_conv_eff: float = 0.95, # power converter efficiency
    # ki: float = 0.01,  # iron losses
    # kw: float = 0.000005,  # windage losses
    # kc: float = 0.3,  # copper losses
    # con_l: int = 600,
) -> float:
    """Battery power modelling.

    Calculate power supplied from or transfer to the battery. 

    Args:
        v (float): velocity (m/s)
        force (float): tractive force (N)
        gear_ratio (float): gear ratio
        radius_wheel (float): dynamic radius of tires (m)
        regen_ratio (float): proportion of braking forcing to use for regen
        trans_eff (float): transmission efficiency
        motor_eff (float): motor efficiency
        pow_conv_eff (float): power converter efficiency
    Returns:
        float: Power supplied from the battery (excluding accessor power)

    """
    
    if force<0:
        force = force*regen_ratio

    torque_wheel = force * radius_wheel #torque at the wheel

    torque_motor = torque_wheel/(gear_ratio* trans_eff) # torque at the motor

    omega_wheel = v/radius_wheel #angular velocity of the wheel
    omega_motor = omega_wheel * gear_ratio #angular velocity of motor

    power = (torque_motor * omega_motor)/(motor_eff*pow_conv_eff) #Power in Watts

    return power

    # TODO include formula for motor efficiency

    # if v==0:
    #     pmot_in=0
    # elif v>0:
    #     if force<0:
    #         force = regen_ratio * force * g_eff
    #     elif force >=0:
    #         force = force / g_eff

    #     torque = (force * rw)/gr
    #     omega = (gr * v)/rw

        # if torque > 0:
        #     eff_mot = (torque * omega) / ((torque * omega) +
        #                                   (np.power(torque, 2) * kc) +
        #                                   (omega * ki) +
        #                                   ((np.power(omega, 3) * kw) + con_l))
        # elif torque < 0:
        #     eff_mot = (-torque * omega) / ((-torque * omega) +
        #                                    (np.power(torque, 2) * kc) +
        #                                    (omega * ki) +
        #                                    ((np.power(omega, 3) * kw) + con_l))

        # if force >= 0:
        #     pmot_in = (force *v) / 0.85
        # elif force < 0:
        #     pmot_in =(force *v) * 0.85

    # return pmot_in

def tractive_force(
    v: float,
    accel: float,
    alpha: Optional[float] = None,
    ch: float = 1.2,
    m: float = 12000.0,
    air_density: float = 1.2,
    cw: float = 1.17,
    area: float = 10.0,
) -> Any:
    """Tractive force.

    Calculation of the force propelling the vehicles forward transferred through the drive wheels.

    Args:
        v (float): velocity (m/s)
        accel (float): acceleration (m/s^2)
        alpha (float): elevation grade
        ch (float): rolling resistance coefficient
        m (float): mass (kg)
        air_density (float): air density
        cw (float): air density coefficient
        area (float): frontal area of vehicle

    Returns:
        float: total tractive force (N)

    """
    g = 9.81

    f = (0.0041 + 0.000041 * v * 2.24) * ch
    roll_resist_force = m * f * g

    if air_density:
        aero_drag_force = 0.5 * air_density * cw * area * np.power(v, 2)
    else:
        aero_drag_force = 0.0

    if alpha:
        hill_climb_force = m * g * np.sin(alpha)
    else:
        hill_climb_force = 0.0

    accel_force = m * accel

    return roll_resist_force + aero_drag_force + hill_climb_force + accel_force


def open_circuit_voltage(x: float, type: str = "LA") -> float:
    """Open Circiut Voltage.

    Li-ion open circuit volage as found in:

    Zhang R. et. al. (2018). A Study on Open Circuit Voltage and State of Charge
    Characterization of High Capacity Lithium-Ion Battery Under Different Temperature.
    Energies 2018, 1, 2408.

    Args:
        x (float): Depth of Discharge
        type (str): battery chemistry type

    Returns:
        float: open circuit voltage


    """

    # TODO: Remove LA and NC battey type since not common anymore

    voltage = 0.0
    
    if type == "LA":
        voltage = (2.15 - ((2.15 - 2.00) * x))
    elif type == "NC":
        voltage = (-8.2816 * (np.power(x, 7)) + 23.5749 *
                               (np.power(x, 6)) - 30 *
                               (np.power(x, 5)) + 23.7053 *
                               (np.power(x, 4)) - 12.5877 * (np.power(x, 3)) +
                               4.1315 * x * x - 0.8658 * x + 1.37)
    elif type == "LI-ION":
        s = 1 - x
        voltage = 0.76 * np.power(s, 5) - 3.72 * np.power(
            s, 4) + 6.15 * np.power(s, 3) - 3.64 * np.power(
                s, 2) + 1.26 * np.power(s, 1) + 3.24

    return voltage



import math
import numpy as np
import numpy.typing as npt
from typing import Any


def const_accel(vi: float = 0.0,
                v_target: float = 13.0,
                vf: float = 0.0,
                di: float = 0.0,
                df: float = 100.0,
                ti: float = 0.0,
                step: float = 0.1,
                a_max: float = 1.0) -> npt.NDArray[Any]:
    """Generate trajectory using constant acceleration.

    Generate linear trajectory with parabolic bends (trapezoidal).
    The trajectory is divided into three parts: the acceleration phase,
    constant velocity and deceleration phase. In some cases, the constant
    velocity phase will not be reached in which case we only have
    a acceleration and deceleration phase.

    Biagiotti, L., Melchiorri, C. (2008). Trajectory Planning for Automatic Machines
    and Robots. Springer-Verlag.

    Args:
        vi (float): initial velocity 
        v_target (float): target velocity 
        vf (float): final velocity 
        di (float): initial position 
        df (float): final position 
        ti (float): initial time
        step (float): time step
        a_max (float): maximum acceleration
    Returns:
        list: (n,3) numpy array of time, velocity, distance of trajectory

    """

    tvq = np.array([(ti, vi, di)])

    t = ti + step

    # We need to first check if the trajectory is even possible given
    # the accel, vi, vf and distance

    assert np.abs(pow(vf, 2) - pow(vi, 2)) / 2 < a_max * (
        df - di), "Error 1: Trajectory is not possible given inputs."
    assert np.sqrt(
        (df - di) / a_max
    ) * 2 > 1, "Error 2: Trapezoidal trajectory is not possible given inputs."

    
    # If the below condition is true, then the target velocity is met 

    if (df - di) * a_max > (pow(v_target, 2) -
                            ((pow(vi, 2) + pow(vf, 2)) / 2)):

        Ta = np.abs(v_target -
                    vi) / a_max  # Use abs if vi is greater than v_target
        Td = (v_target - vf) / a_max
        T = ti + ((df - di) / v_target) \
            + ((v_target / (2 * a_max)) * pow((1 - (vi / v_target)), 2)) \
            + ((v_target / (2 * a_max)) * pow((1 - (vf / v_target)), 2))

        # print('Case 1')
        # print(f"Ta: {Ta}, Td: {Td}, T: {T}")

        while t <= T:
            if t >= ti and t < ti + Ta and Ta != 0:  # Acceleration phase
                v = _accel_v(vi, v_target, Ta, t, ti)
                q = _accel_q(di, vi, ti, v_target, Ta, t)
            elif t >= ti + Ta and t < T - Td:  # Constant velocity phase
                v = v_target
                q = _const_q(di, vi, Ta, v_target, ti, t)
            elif t >= T - Td and t <= T and Td != 0:  # Deceleration phase
                v = _decel_v(vf, v_target, Td, T, t)
                q = _decel_q(df, vf, T, t, v_target, Td)
            elif Ta == 0:  # case when Vi and v_target are equal
                v = v_target
                q = di + (v_target * t)
            else:
                raise ValueError(
                    'Cannot create trajectory using input values.')

            s = np.column_stack((t, v, q))
            tvq = np.concatenate([tvq, s])
            t += step

    else:
        vlim = math.sqrt(((df - di) * a_max) + ((pow(vi, 2) + pow(vf, 2)) / 2))
        Ta = (vlim - vi) / a_max
        Td = (vlim - vf) / a_max
        T = ti + Ta + Td

        # print('Case 2')
        # print(f"Ta: {Ta}, Td: {Td}, T: {T}, vlim: {vlim}")

        while t <= T:
            if t >= ti and t < ti + Ta:
                v = _accel_v(vi, vlim, Ta, t, ti)
                q = _accel_q(di, vi, ti, vlim, Ta, t)
            elif t >= ti + Ta and t <= T and Td != 0:
                v = _decel_v(vf, vlim, Td, T, t)
                q = _decel_q(df, vf, T, t, vlim, Td)
            elif Td == 0:
                v = vf
                q = di + (df * t)
            else:
                raise ValueError(
                    'Cannot create trajectory using input values.')

            s = np.column_stack((t, v, q))
            tvq = np.concatenate([tvq, s])
            t += step

    return tvq 
    
    # TODO: need a different data structure for the time, velocit, data return. 
    # Maybe a dictionary


def _accel_q(di: float, vi: float, ti: float, v_target: float, Ta: float,
             t: float) -> float:
    return di + (vi * (t - ti)) + (((v_target - vi) / (2 * Ta)) * pow(
        (t - ti), 2))


def _accel_v(vi: float, v_target: float, Ta: float, t: float,
             ti: float) -> float:
    return vi + ((v_target - vi) / Ta) * (t - ti)


def _const_q(di: float, vi: float, Ta: float, v_target: float, ti: float,
             t: float) -> float:
    return di + (vi * (Ta / 2)) + (v_target * (t - ti - (Ta / 2)))


def _decel_q(df: float, vf: float, T: float, t: float, v_target: float,
             Td: float) -> float:
    return df - (vf * (T - t)) - (((v_target - vf) / (2 * Td)) * pow(
        (T - t), 2))


def _decel_v(vf: float, v_target: float, Td: float, T: float,
             t: float) -> float:
    return vf + ((v_target - vf) / Td) * (T - t)



import networkx as nx
from typing import List, Dict, Any, Tuple


class Graph:
    """Generate a path graph of a bus route

    Generate a path graph of a bus route using a list of dicts of edge
    attributed such as end node, speed, and linear reference. The purpose
    of this class is to simplify paths by clustering nearby intersections.

    Attributes:
        graph: Networkx graph of route
        source: First node of the graph
        target: Last node of the graph
        edges: List of Dict of edges

    """

    def __init__(self, edges: List[Dict[str, Any]]):

        self.graph = self.dict_to_graph(edges)
        self.source = 0
        self.target = len(edges)
        self.edges = edges

    def get_source_target(self) -> Tuple[int, int]:
        return (self.source, self.target)

    @staticmethod
    def dict_to_graph(data_: List[Dict[str, Any]]) -> nx.Graph:

        G = nx.path_graph(len(data_) + 1, nx.DiGraph())
        keys = list(data_[0].keys())

        x = 0
        intersection = [None]

        for i, (u, v) in enumerate(G.edges):
            for key in keys:
                G[u][v][key] = data_[i][key]

            G.nodes[u]["intersection"] = intersection
            G.nodes[u]["LR"] = x

            G.nodes[v]["intersection"] = data_[i]["intersection"]
            G.nodes[v]["LR"] = x + data_[i]["length"]

            x = G.nodes[v]["LR"]
            intersection = G.nodes[v]["intersection"]

        return G

    def get_edges(self) -> List[Dict[str, Any]]:
        """Get list of edges from networkx DiGraph.

        Returns:
            list: List of dictionaries of route path edges.

        """
        edges = []

        for u, v in nx.bfs_edges(self.graph, self.source):
            edges.append(self.graph[u][v])

        return edges

    def consolidate_intersections(
        self,
        filters: List[str] = ["tertiary", "secondary",
                              "bus_stop"]) -> nx.Graph:
        """Consolidate intersections that are clustered together

        Args:
            filters (list): list intersections to be clustered.
        Returns:
            Networkx DiGraph
        """

        H = self.graph.copy()

        for u, v in nx.bfs_successors(self.graph, 0):

            n = v[0]

            in_edge = list(H.in_edges(n))
            out_edge = list(H.out_edges(n))

            if len(in_edge) > 0 and len(out_edge) > 0:

                u = in_edge[0]
                v = out_edge[0]

                e1 = H.edges[u]
                e2 = H.edges[v]

                l1 = e1["length"]
                l2 = e2["length"]

                if l1 < 100 and any(x in H.nodes[n]["intersection"]
                                    for x in filters) and any(
                                        x in H.nodes[u[0]]["intersection"]
                                        for x in filters):
                    # print(H.nodes[u[0]]["intersection"],H.nodes[n]["intersection"])
                    H.add_edge(u[0],
                               v[1],
                               way_id=e1["way_id"] + e2["way_id"],
                               speed=e1["speed"],
                               length=l1 + l2,
                               intersection=H.nodes[v[1]]["intersection"])
                    H.nodes[u[0]]["intersection"] = H.nodes[n][
                        "intersection"] + H.nodes[u[0]]["intersection"]
                    H.remove_node(n)

                elif l2 < 100 and any(x in H.nodes[n]["intersection"]
                                      for x in filters) and any(
                                          x in H.nodes[v[1]]["intersection"]
                                          for x in filters):
                    # print(H.nodes[v[1]]["intersection"],H.nodes[n]["intersection"])
                    H.add_edge(u[0],
                               v[1],
                               way_id=e1["way_id"] + e2["way_id"],
                               speed=e1["speed"],
                               length=l1 + l2,
                               intersection=H.nodes[v[1]]["intersection"] +
                               H.nodes[n]["intersection"])
                    H.nodes[v[1]]["intersection"] = H.nodes[n][
                        "intersection"] + H.nodes[v[1]]["intersection"]
                    H.remove_node(n)

                else:
                    continue

        self.graph = H
        return H

    def simplify_graph(
        self,
        filters: List[str] = ["tertiary", "secondary", "bus_stop"]
    ) -> 'nx.Graph[Any]':
        """Method to simplify graph by clustering adjacent nodes

        This method clusters adjacent nodes that are included in the filters list
        and that have adjacent edges that have the same speed.

        Args:
            filters (list): list of intersections that should NOT be merged
        Returns:
            Networkx DiGraph
        """

        H = self.graph.copy()

        for u, v in nx.bfs_successors(self.graph, 0):

            n = v[0]

            in_edge = list(H.in_edges(n))
            out_edge = list(H.out_edges(n))

            if len(in_edge) > 0 and len(out_edge) > 0:

                u = in_edge[0]
                v = out_edge[0]

                e1 = H.edges[u]
                e2 = H.edges[v]

                if e1["speed"] == e2["speed"] and \
                    not any(x in e1["intersection"] for x in filters):

                    sum_length = e1["length"] + e2["length"]
                    speed = e1["speed"]
                    way_ids = e1["way_id"] + e2["way_id"]
                    intersections = e1["intersection"] + e2["intersection"]

                    H.add_edge(u[0],
                               v[1],
                               way_id=way_ids,
                               speed=speed,
                               length=sum_length,
                               intersection=intersections)
                    H.remove_node(n)

        self.graph = H
        return H

    def include_stops(self, stops: List[float]) -> 'nx.Graph[Any]':
        """Method to insert bus stop location in networkx DiGraph

        Args:
            stops (list): List of linear referenced location of bus stops
        Returns:
            Networkx DiGraph

        """
        J = self.graph.copy()

        for u, v in nx.bfs_edges(self.graph, 0):
            di = self.graph.nodes[u]["LR"]
            df = self.graph.nodes[v]["LR"]

            bus_stops = [i for i in stops if i >= di and i < df]

            if len(bus_stops) > 0:

                for stop in bus_stops:
                    J.add_node(stop, intersection=["bus_stop"], LR=stop)

                stops1 = [u, *bus_stops, v]

                for i in range(len(stops1) - 1):
                    J.add_edge(stops1[i],
                               stops1[i + 1],
                               way_id=self.graph[u][v]["way_id"],
                               speed=self.graph[u][v]["speed"],
                               length=J.nodes[stops1[i + 1]]["LR"] -
                               J.nodes[stops1[i]]["LR"],
                               intersection=J.nodes[stops1[i +
                                                           1]]["intersection"])
                J.remove_edge(u, v)

        self.graph = J
        return J



import numpy as np
from drivecycle import trajectory

from typing import List, Dict, Any
# import numpy.typing as npt

import logging

logging.getLogger().setLevel(logging.INFO)


def sequential(edges: List[Dict[str, Any]],
               stops: Dict[str, int],
               di: float = 0,
               vi: float = 0,
               ti: float = 0,
               a_max: float = 1,
               step: float = 0.1,
               stop_at_node: bool = False,
               kmh: bool = True) -> 'np.ndarray':  # type: ignore
    """Generate route drivecycle.

    Generate a drivecycle which is formed from a collection or 
    sequence of trajectories. 

    Args:
        edges (list): list of dictionary of route edges
        stops (dict): dict of stops categories and stop durations
        di (float): initial position
        vi (float): initial velocity
        ti (float): initial time
        a_max (float): maximum acceleration
        step (float): time step
        kmh (bool): kilometers per hour

    Returns:
        list: (n,3) numpy array of time, velocity, distance of drivecycle

    """

    tvq = np.array([(ti, vi, di)])

    stop = None
    conversion = 1000 / 3600

    keys = list(stops.keys())

    a = a_max

    carryoverdistance = None

    for i in range(len(edges)):

        if kmh:
            conversion = 1000 / 3600
        else:
            conversion = 1

        try:
            v_target_next = edges[i + 1]["speed"] * conversion
        except IndexError:
            v_target_next = 0

        v_target = edges[i]["speed"] * conversion
        if carryoverdistance:
            df = di + edges[i]["length"]+carryoverdistance
            carryoverdistance=None
        else:
            df = di + edges[i]["length"]

        if any(x in list(stops.keys())
               for x in edges[i]["intersection"]) and stop_at_node:
            stop = np.random.randint(2)

        if stop:
            vf = 0
            a = 1
        else:
            if v_target_next >= v_target:
                vf = v_target
            else:
                vf = v_target_next

        while True:
            try:
                d = trajectory.const_accel(vi=vi,
                                        v_target=v_target,
                                        vf=vf,
                                        di=di,
                                        df=df,
                                        ti=ti,
                                        step=step,
                                        a_max=a)
                break
            except AssertionError: # if the trajectory segment is not feasible then try the following:
                if v_target>=1.0 and vf>=1.0: # (1) reduce v_target and vf by 10% if they are not both less than 1 m/s
                    v_target = v_target*.9
                    vf = vf*.9
                    logging.info(f"Vi: {vi:.2f}. Reducing vf to {vf:.2f} and v_target to {v_target:.2f} at time {ti} and segment length {df-di}")
                else: 
                    if vi!=0.0: # (2) if that doesn't work then use a constant velocity accross the segment
                        tf = (df - di)/vi  # Constant veclocity using vi
                        d = np.array([[ti, vi, di], [ti+tf, vi, df]])
                        logging.info(
                            f'Could not complete segment: ti: {ti:.2f}, tf: {tf:.2f},  vi: {vi:.2f} , vf: {vf:.2f}, \
                                v_target:{v_target:.2f}, length: {edges[i]["length"]:.2f}')
                        break
                    else: # (3) if vi==0 then can't complete this segment. Carry over the distance for this segment to the next segment
                        d = np.array([[ti, vi, di]])
                        carryoverdistance = df-di
                        logging.info('Carry over segment length to next segment')
                        break


        if stop:
            stop_id = next(x for x in keys if x in edges[i]["intersection"])
            stop_max_time = stops[stop_id]
            stop_time = np.random.randint(5, stop_max_time)

            t = np.linspace(d[-1][0] + step, d[-1][0] + stop_time, 5)
            v = np.zeros(5)
            q = np.repeat(d[-1][2], 5)
            s = np.column_stack((t, v, q))
            d = np.concatenate([d, s])

        di = d[-1][2]
        ti = d[-1][0]
        vi = d[-1][1]

        tvq = np.concatenate([tvq, d[:-1]])

        stop = 0

    return tvq[1:]  # type: ignore



import matplotlib.pyplot as plt
from typing import Any, List

plt.style.use('ggplot')


def plot(v: Any,
         i: Any,
         savepath: str,
         labels: List[str] = ["velocity", "time"]) -> None:
    plt.rcParams['savefig.facecolor'] = 'white'
    plt.rcParams.update({'font.size': 14})
    plt.figure(figsize=(15, 8))
    plt.plot(i, v, color='#4b0082', linewidth=1.5)
    plt.xlabel(labels[1])
    plt.ylabel(labels[0])

    if savepath:
        plt.savefig(savepath, transparent=False, bbox_inches='tight')

    plt.show()


def plot_vt(tvq: Any, savepath: str) -> None:
    t = [i[0] for i in tvq]
    v = [i[1] for i in tvq]

    plot(v, t, savepath, labels=["velocity (m/s)", "time (s)"])


def plot_vd(tvq: Any, savepath: str) -> None:
    d = [i[2] for i in tvq]
    v = [i[1] for i in tvq]

    plot(v, d, savepath, labels=["velocity (m/s)", "distance (m)"])


def plot_dt(tvq: Any, savepath: str) -> None:
    d = [i[2] for i in tvq]
    t = [i[0] for i in tvq]

    plot(d, t, savepath, labels=["distance (m)", "time (s)"])
